/*
 * Copyright (c) 2001, Swedish Institute of Computer Science.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 * $Id: tcp.c,v 1.10 2001/01/31 10:57:30 adam Exp $
 */

/*-----------------------------------------------------------------------------------*/
/* tcp.c
 *
 * This file contains common functions for the TCP implementation, such as functinos
 * for manipulatong the data structures and the TCP timer functions. TCP functions
 * related to input and output is found in tcp_input.c and tcp_output.c respectively.
 *
 */
/*-----------------------------------------------------------------------------------*/

#include "debug.h"

#include "def.h"
#include "mem.h"

#include "tcp.h"

/* The list of TCP PCBs. */
struct tcp_pcb *tcp_pcbs;

/* incremented every fine grained timer shot
   (typically every 200 ms, determined by TCP_FINE_TIMEOUT) */
u32_t tcp_ticks;

#define MIN(x,y) (x) < (y)? (x): (y)
/*-----------------------------------------------------------------------------------*/
/* tcp_close:
 *
 * Closes the connection held by the PCB. It sends a FIN segment to the remote host
 * and enters the FIN-WAIT-1 state.
 */
/*-----------------------------------------------------------------------------------*/
int
tcp_close(struct tcp_pcb *pcb)
{
  /*  DEBUGF("tcp_close:\n");*/
  switch(pcb->state) {
  case CLOSED:
  case LISTEN:
  case SYN_SENT:
    tcp_pcb_remove(pcb);
    break;
  case SYN_RCVD:
    tcp_send_ctrl(pcb, TCP_FIN);
    pcb->state = FIN_WAIT_1;
    break;
  case ESTABLISHED:    
    tcp_send_ctrl(pcb, TCP_FIN);
    pcb->state = FIN_WAIT_1;
    break;
  case CLOSE_WAIT:
    tcp_send_ctrl(pcb, TCP_FIN);
    pcb->state = LAST_ACK;
    break;
  default:
    /* Has already been closed, do nothing. */
    break;
  }
  return 0;
}
/*-----------------------------------------------------------------------------------*/
void
tcp_register(struct tcp_pcb *pcb)
{
  struct tcp_pcb *ipcb;
  for(ipcb = tcp_pcbs; ipcb != NULL; ipcb = ipcb->next) {
    if(pcb == ipcb) {
      DEBUGF(TCP_DEBUG, ("tcp_register: already registered\n"));
      return;
    }
  }
  ASSERT("tcp_register: pcb->state != CLOSED", pcb->state != CLOSED);
  pcb->next = tcp_pcbs;
  tcp_pcbs = pcb;
}
/*-----------------------------------------------------------------------------------*/
int
tcp_bind(struct tcp_pcb *pcb, struct ip_addr *ipaddr, u16_t port)
{
  if(ipaddr != NULL) {
    pcb->local_ip = *ipaddr;
  }
  pcb->local_port = port;
  DEBUGF(TCP_DEBUG, ("tcp_bind: bind to port %d\n", port));
  return 0;
}
/*-----------------------------------------------------------------------------------*/
int
tcp_listen(struct tcp_pcb *pcb)
{
  pcb->state = LISTEN;
  mem_realloc(pcb, sizeof(struct tcp_pcb_listen));
  tcp_register(pcb);
  return 0;
}
/*-----------------------------------------------------------------------------------*/
void
tcp_recved(struct tcp_pcb *pcb, u16_t len)
{
  pcb->rcv_wnd += len;
  if(pcb->rcv_wnd > TCP_WND) {
    pcb->rcv_wnd = TCP_WND;
  }
  DEBUGF(TCP_DEBUG, ("tcp_recved: recveived %d bytes, wnd %lu (%lu).\n",
		     len, pcb->rcv_wnd, TCP_WND - pcb->rcv_wnd));
}
/*-----------------------------------------------------------------------------------*/
/* tcp_new_port():
 *
 * A nastly hack using a 'goto' statement to allocate a new TCP local port.
 */
/*-----------------------------------------------------------------------------------*/
static unsigned short 
tcp_new_port(void)
{
  struct tcp_pcb *pcb;
  static u16_t port = 4096;
 again:
  if(++port > 32000) {
    port = 4096;
  }
  
  for(pcb = tcp_pcbs; pcb != NULL; pcb = pcb->next) {
    if(pcb->local_port == port) {
      goto again;
    }
  }
  return port;
}
/*-----------------------------------------------------------------------------------*/
int
tcp_connect(struct tcp_pcb *pcb, struct ip_addr *ipaddr, u16_t port)
{
  char optdata[4];
  u32_t iss;

  DEBUGF(TCP_DEBUG, ("tcp_connect to port %d\n", port));
  if(ipaddr != NULL) {
    pcb->dest_ip = *ipaddr;
  } else {
    return -1;
  }
  pcb->dest_port = port;
  pcb->local_port = tcp_new_port();
  iss = tcp_next_iss();
  pcb->rcv_nxt = 0;
  pcb->snd_nxt = iss;
  pcb->snd_ack = iss - 1;
  pcb->snd_lbb = iss - 1;
  pcb->rcv_wnd = TCP_WND;
  pcb->snd_wnd = TCP_WND;
  pcb->mss = TCP_MSS;
  pcb->cwnd = 1;
  pcb->ssthresh = pcb->mss * 10;
  pcb->state = SYN_SENT;
  tcp_register(pcb);  
  /* Build an MSS option */
  optdata[0] = 2;   /* option kind (MSS) */
  optdata[1] = 4;   /* option length */
  optdata[2] = pcb->mss / 256;
  optdata[3] = pcb->mss & 255;
  tcp_enqueue(pcb, NULL, 0, TCP_SYN, 0, optdata, 4);
  return 0;
} 
/*-----------------------------------------------------------------------------------*/
/* tcp_timer_coarse:
 *
 * Called every 500 ms and implements the retransmission timer and the timer that
 * removes PCBs that have been in TIME-WAIT for enough time. It also increments
 * various timers such as the inactivity timer in each PCB.
 */
/*-----------------------------------------------------------------------------------*/
static void
tcp_timer_coarse(void *data)
{
  struct tcp_pcb *pcb, *pcb2, *prev;
  struct tcp_seg *seg;
  u32_t eff_wnd;
  u8_t pcb_remove;      /* flag if a PCB should be removed */

  tcp_ticks++;
    
  prev = NULL;
  /* Steps through all of the PCBs. */
  pcb = tcp_pcbs;
  while(pcb != NULL) {
    ASSERT("tcp_timer_coarse: pcb->state != CLOSED", pcb->state != CLOSED);
    
    if(pcb->state != TIME_WAIT &&
       pcb->state != LISTEN) {
      /* check if we need to do any retransmissions */
      for(seg = pcb->unacked; seg != NULL; seg = seg->next) {
	seg->rtime++;
	if(seg->rtime >= pcb->rto) {

	  DEBUGF(TCP_RTO_DEBUG, ("tcp_timer_coarse: seg->rtime %d pcb->rto %d\n",
		 seg->rtime, pcb->rto));

	  tcp_rexmit_seg(pcb, seg);
	  eff_wnd = MIN(pcb->cwnd, pcb->snd_wnd);
	  pcb->ssthresh = eff_wnd >> 1;
	  if(pcb->ssthresh < pcb->mss) {
	    pcb->ssthresh = pcb->mss * 2;
	  }
	  pcb->cwnd = pcb->mss;

	  DEBUGF(TCP_CWND_DEBUG, ("tcp_rexmit_seg: cwnd %lu ssthresh %lu\n", pcb->cwnd,
		 pcb->ssthresh));
	}
      }
      if(pcb->state == FIN_WAIT_2 && pcb->tmr < 0xffff) {
	pcb->tmr++;
      }

    }
    
    pcb_remove = 0;

    /* Check if this PCB has stayed long enough in TIME-WAIT */
    if(pcb->state == TIME_WAIT) {
      pcb->tmr++;
      if(pcb->tmr > 2 * TCP_MSL / TCP_COARSE_TIMEOUT) {
        pcb_remove++;
      }
    }    


    /* If the PCB should be removed, do it. */
    if(pcb_remove) {
      tcp_pcb_purge(pcb);      
      /* Remove PCB from tcp_pcbs list. */
      if(prev != NULL) {
	ASSERT("tcp_timer_coarse: middle tcp != tcp_pcbs", pcb != tcp_pcbs);
        prev->next = pcb->next;
      } else {
        /* This PCB was the first. */
        ASSERT("tcp_timer_coarse: first pcb == tcp_pcbs", tcp_pcbs == pcb);
        tcp_pcbs = pcb->next;
      }
      pcb2 = pcb->next;
      mem_free(pcb);
      pcb = pcb2;
    } else {
      prev = pcb;
      pcb = pcb->next;
    }
  }
  sys_timeout(TCP_COARSE_TIMEOUT, (sys_timeout_handler)tcp_timer_coarse, NULL);
}
/*-----------------------------------------------------------------------------------*/
/* tcp_timer_fine:
 *
 * Is called every TCP_FINE_TIMEOUT (100 ms) and sends delayed ACKs.
 */
/*-----------------------------------------------------------------------------------*/
static void
tcp_timer_fine(void *data)
{
  struct tcp_pcb *pcb;
  
  for(pcb = tcp_pcbs; pcb != NULL; pcb = pcb->next) {

    /* Increment timers. */
    /*    if(pcb->rttest) {
      pcb->rttest++;
      }*/

    /* send delayed ACKs */
    if(pcb->state != LISTEN &&
       pcb->state != TIME_WAIT &&
       pcb->flags & TCP_ACK_NEXT) {
      pcb->flags &= ~TCP_ACK_NEXT;
      tcp_send_ctrl(pcb, TCP_ACK);
      DEBUGF(TCP_DEBUG, ("tcp_timer_fine: delayed ACK\n"));
    }
  }
  
  sys_timeout(TCP_FINE_TIMEOUT, (sys_timeout_handler)tcp_timer_fine, NULL);
}
/*-----------------------------------------------------------------------------------*/
/* tcp_segs_free:
 *
 * Deallocates a list of TCP segments (tcp_seg structures).
 *
 * It returns as approximation of the number of freed bytes.
 */
/*-----------------------------------------------------------------------------------*/
unsigned int
tcp_segs_free(struct tcp_seg *seg)
{
  if(seg != NULL) {
    return tcp_segs_free(seg->next) + tcp_seg_free(seg);
  }
  return 0;
}
/*-----------------------------------------------------------------------------------*/
/* tcp_seg_free:
 *
 * Frees a TCP segment and returns the approximate number of memory bytes freed.
 *
 * An important note is that the segment might be buffered somewhere in the system,
 * and should therefore not be removed. The call to pbuf_free returns with 0 if the
 * pbuf could be freed (i.e. if no one else was using it), and then the segment can
 * also be freed. If pbuf_ref returns non-zero, the segment is in use and can
 * therefore not be deallocated.
 */
/*-----------------------------------------------------------------------------------*/
unsigned int
tcp_seg_free(struct tcp_seg *seg)
{
  unsigned int size;
  if(seg != NULL) {
    if(seg->p == NULL) {
      mem_free(seg);
      return 0;
    } else {
      size = seg->p->tot_len;
      
      if(pbuf_free(seg->p) == 0) {
	/*	DEBUGF("tcp_seg_free: didn't free 0x%lx\n", (long)seg);*/
	return 0;
      } else {
	mem_free(seg);
	return size;
      }
    }
  }
  return 0;
}
/*-----------------------------------------------------------------------------------*/
struct tcp_pcb *
tcp_pcb_new(void)
{
  struct tcp_pcb *pcb;
  pcb = mem_malloc(sizeof(struct tcp_pcb));
  if(pcb != NULL) {
    bzero(pcb, sizeof(struct tcp_pcb));
    pcb->unsent = pcb->unacked = pcb->ooseq = NULL;
    pcb->rcv_wnd = TCP_WND;
    pcb->mss = TCP_MSS;
    pcb->rto = 3000 / TCP_COARSE_TIMEOUT;
    pcb->sa = 0;
    pcb->sv = 3000 / TCP_COARSE_TIMEOUT;
    return pcb;
  }
  return NULL;
}
/*-----------------------------------------------------------------------------------*/
/* tcp_mem_reclaim:
 *
 * Tries to free up TCP memory. This function is called from the memory manager
 * when memory is scarce.
 */
/*-----------------------------------------------------------------------------------*/
#ifdef MEM_RECLAIM
static unsigned int
tcp_mem_reclaim(void *data, unsigned int size)
{
  struct tcp_pcb *pcb, *npcb, *inactive;
  unsigned int inactivity, reclaimed;
  unsigned long seqno, ackno;
  unsigned short local_port, dest_port;
  struct ip_addr local_ip, dest_ip;
  char removed;
  
  reclaimed = 0;  
  
  /* First, try to free any out of sequence data in the PCBs.
     Also kill any connections in TIME-WAIT. Next, kill the PCB
     with the longest inactivity time (unless they are in LISTEN state).
     The two loops are combined. */

  inactivity = 0;
  inactive = NULL;
  pcb = tcp_pcbs;
  while(reclaimed < size && pcb != NULL) {
    removed = 0;
    if(pcb->ooseq) {
      reclaimed += tcp_segs_free(pcb->ooseq);
      DEBUGF(TCP_DEBUG, ("tcp_mem_reclaim: reclaiming memory from PCB 0x%lx\n", (long)pcb));
    }
    if(pcb->state == TIME_WAIT) {
      npcb = pcb->next;
      tcp_pcb_remove(pcb);
      removed = 1;
    }
    /*    if(pcb->state != LISTEN &&
       pcb->state != TIME_WAIT &&
       pcb->timer_inactive > inactivity) {
      inactive = pcb;
      inactivity = pcb->timer_inactive;
      }*/
    if(!removed) {
      pcb = pcb->next;
    }
  }
  
  if(reclaimed < size && inactive != NULL) {
    DEBUGF(TCP_DEBUG, ("tcp_mem_reclaim: killing oldest PCB 0x%lx (%d)\n",
		       (long)inactive, inactivity));
    seqno = inactive->snd_nxt;
    ackno = inactive->rcv_nxt;
    ip_addr_set(&local_ip, &(inactive->local_ip));
    ip_addr_set(&dest_ip, &(inactive->dest_ip));
    local_port = inactive->local_port;
    dest_port = inactive->dest_port;
    reclaimed += tcp_pcb_remove(inactive);
    tcp_rst(seqno, ackno, &local_ip, &dest_ip, local_port, dest_port); 
  }
  

  DEBUGF(TCP_DEBUG, ("tcp_mem_reclaim: reclaimed %d bytes (wanted %d)\n", reclaimed, size));
  return reclaimed;
}
#endif /* MEM_RECLAIM */
/*-----------------------------------------------------------------------------------*/
/* tcp_init:
 *
 * Initializes the TCP layer.
 */
/*-----------------------------------------------------------------------------------*/
void
tcp_init(void)
{
  /* initialize timers */
  tcp_ticks = 0;
  sys_timeout(TCP_FINE_TIMEOUT, (sys_timeout_handler)tcp_timer_fine, NULL);
  sys_timeout(TCP_COARSE_TIMEOUT, (sys_timeout_handler)tcp_timer_coarse, NULL);

#ifdef MEM_RECLAIM
  /* Register memory reclaim function */
  mem_register_reclaim(tcp_mem_reclaim, NULL);
#endif /* MEM_RECLAIM */
}
/*-----------------------------------------------------------------------------------*/
void
tcp_recv(struct tcp_pcb *pcb,
	 void (* recv)(void *arg, struct tcp_pcb *pcb, struct pbuf *p),
	 void *recv_arg)
{
  pcb->recv = recv;
  pcb->recv_arg = recv_arg;
}
/*-----------------------------------------------------------------------------------*/
void
tcp_accept(struct tcp_pcb *pcb,
	   void (* accept)(void *arg, struct tcp_pcb *newpcb),
	   void *accept_arg)
{
  pcb->accept = accept;
  pcb->accept_arg = accept_arg;
}
/*-----------------------------------------------------------------------------------*/
/* tcp_pcb_purge:
 *
 * Purges a TCP PCB. Removes any buffered data and frees the buffer memory.
 */
/*-----------------------------------------------------------------------------------*/
unsigned int 
tcp_pcb_purge(struct tcp_pcb *pcb)
{
  unsigned int size;
  if(pcb->state != TIME_WAIT &&
     pcb->state != LISTEN) {

#if TCP_DEBUG
    if(pcb->unsent != NULL) {    
      DEBUGF(TCP_DEBUG, ("tcp_pcb_purge: not all data sent\n"));
    }
#endif /* TCP_DEBUG */
    size = tcp_segs_free(pcb->unsent) + 
      tcp_segs_free(pcb->unacked) +
      tcp_segs_free(pcb->ooseq);
    pcb->unacked = pcb->unsent = pcb->ooseq = NULL;
    return size;
  }
  return 0;
}
/*-----------------------------------------------------------------------------------*/
/* tcp_pcb_remove:
 *
 * Purges the PCB and removes it from the PCB list. Any delayed ACKs are sent first.
 */
/*-----------------------------------------------------------------------------------*/
unsigned int
tcp_pcb_remove(struct tcp_pcb *pcb)
{
  unsigned int size;
  struct tcp_pcb *pcb2;

  /* if there is an outstanding delayed ACKs, send it */
  if(pcb->state != TIME_WAIT &&
     pcb->state != LISTEN &&
     pcb->flags & TCP_ACK_NEXT) {
    tcp_send_ctrl(pcb, TCP_ACK);
  }
  
  if(tcp_pcbs == pcb) {
    tcp_pcbs = tcp_pcbs->next;
  } else for(pcb2 = tcp_pcbs; pcb2 != NULL; pcb2 = pcb2->next) {
    if(pcb2->next != NULL && pcb2->next == pcb) {
      pcb2->next = pcb->next;
      break;
    }
  }

  size = tcp_pcb_purge(pcb);
  mem_free(pcb);

  return size;
}
/*-----------------------------------------------------------------------------------*/
/* tcp_next_iss:
 *
 * Calculates a new initial sequence number for new connections.
 */
/*-----------------------------------------------------------------------------------*/
u32_t
tcp_next_iss(void)
{
  static u32_t iss = 6510;
  
  iss += tcp_ticks;       /* XXX */
  return iss;
}
/*-----------------------------------------------------------------------------------*/
#if TCP_DEBUG
void
tcp_debug_print(struct tcp_hdr *tcphdr)
{
  DEBUGF(TCP_DEBUG, ("TCP header:\n"));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(TCP_DEBUG, ("|     %5d     |     %5d     | (src port, dest port)\n",
		     tcphdr->src, tcphdr->dest));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(TCP_DEBUG, ("|            %08lu           | (seq no)\n",
			    tcphdr->seqno));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(TCP_DEBUG, ("|            %08lu           | (ack no)\n",
		     tcphdr->ackno));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
#ifdef HAVE_BITFIELDS
  DEBUGF(TCP_DEBUG, ("| %2d |    |%d%d%d%d%d|    %5d      | (offset, flags (",
		     tcphdr->offset,
		     tcphdr->flags >> 4 & 1,
		     tcphdr->flags >> 3 & 1,
		     tcphdr->flags >> 2 & 1,
		     tcphdr->flags >> 1 & 1,
		     tcphdr->flags & 1,
		     tcphdr->wnd));
#else
  DEBUGF(TCP_DEBUG, ("| %2d |    |%d%d%d%d%d|    %5d      | (offset, flags (",
		     tcphdr->flags >> 4 & 1,
		     tcphdr->flags >> 4 & 1,
		     tcphdr->flags >> 3 & 1,
		     tcphdr->flags >> 2 & 1,
		     tcphdr->flags >> 1 & 1,
		     tcphdr->flags & 1,
		     tcphdr->wnd));
#endif /* HAVE_BITFIELDS */
  tcp_debug_print_flags(tcphdr->flags);
  DEBUGF(TCP_DEBUG, ("), win)\n"));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(TCP_DEBUG, ("|    0x%04x     |     %5d     | (chksum, urgp)\n",
		     ntohs(tcphdr->chksum), ntohs(tcphdr->urgp)));
  DEBUGF(TCP_DEBUG, ("+-------------------------------+\n"));
}
/*-----------------------------------------------------------------------------------*/
void
tcp_debug_print_state(enum tcp_state s)
{
  DEBUGF(TCP_DEBUG, ("State: "));
  switch(s) {
  case CLOSED:
    DEBUGF(TCP_DEBUG, ("CLOSED\n"));
    break;
 case LISTEN:
   DEBUGF(TCP_DEBUG, ("LISTEN\n"));
   break;
  case SYN_SENT:
    DEBUGF(TCP_DEBUG, ("SYN_SENT\n"));
    break;
  case SYN_RCVD:
    DEBUGF(TCP_DEBUG, ("SYN_RCVD\n"));
    break;
  case ESTABLISHED:
    DEBUGF(TCP_DEBUG, ("ESTABLISHED\n"));
    break;
  case FIN_WAIT_1:
    DEBUGF(TCP_DEBUG, ("FIN_WAIT_1\n"));
    break;
  case FIN_WAIT_2:
    DEBUGF(TCP_DEBUG, ("FIN_WAIT_2\n"));
    break;
  case CLOSE_WAIT:
    DEBUGF(TCP_DEBUG, ("CLOSE_WAIT\n"));
    break;
  case CLOSING:
    DEBUGF(TCP_DEBUG, ("CLOSING\n"));
    break;
  case LAST_ACK:
    DEBUGF(TCP_DEBUG, ("LAST_ACK\n"));
    break;
  case TIME_WAIT:
    DEBUGF(TCP_DEBUG, ("TIME_WAIT\n"));
   break;
  }
}
/*-----------------------------------------------------------------------------------*/
void
tcp_debug_print_flags(u8_t flags)
{
  if(flags & TCP_FIN) {
    DEBUGF(TCP_DEBUG, ("FIN "));
  }
  if(flags & TCP_SYN) {
    DEBUGF(TCP_DEBUG, ("SYN "));
  }
  if(flags & TCP_RST) {
    DEBUGF(TCP_DEBUG, ("RST "));
  }
  if(flags & TCP_PSH) {
    DEBUGF(TCP_DEBUG, ("PSH "));
  }
  if(flags & TCP_ACK) {
    DEBUGF(TCP_DEBUG, ("ACK "));
  }
  if(flags & TCP_URG) {
    DEBUGF(TCP_DEBUG, ("URG "));
  }
}
/*-----------------------------------------------------------------------------------*/
void
tcp_debug_print_pcbs(void)
{
  struct tcp_pcb *pcb;
  DEBUGF(TCP_DEBUG, ("PCB states:\n"));
  for(pcb = tcp_pcbs; pcb != NULL; pcb = pcb->next) {
    DEBUGF(TCP_DEBUG, ("Local port %d, foreign port %d snd_nxt %lu rcv_nxt %lu ",
                       pcb->local_port, pcb->dest_port,
                       pcb->snd_nxt, pcb->rcv_nxt));
    tcp_debug_print_state(pcb->state);
  }    
}
#endif /* TCP_DEBUG */
/*-----------------------------------------------------------------------------------*/









